# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt
from ikarus.generator import MySimpleGenerator


def boundaryPatch(gridView, booleanVector):
    """
    @brief Creates a boundary patch for the given grid view and boolean vector.

    @param gridView: The grid view.
    @param booleanVector: The boolean vector.

    @return: The created boundary patch.
    """
    generator = MySimpleGenerator("BoundaryPatch", "Ikarus::Python")
    element_type = f"BoundaryPatch<{gridView.cppTypeName}>"

    includes = []
    includes += ["dune/fufem/boundarypatch.hh"]
    includes += ["ikarus/python/utils/boundarypatch.hh"]
    includes += gridView._includes
    moduleName = "boundaryPatch_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.BoundaryPatch(gridView, booleanVector)


from io import StringIO
from dune.generator.algorithm import run
from dune.common import FieldVector


def globalIndexFromGlobalPosition(basis, pos):

    runCode = """
        #include <ikarus/utils/functionhelper.hh>
        #include <dune/python/pybind11/pybind11.h>
        #include <dune/python/pybind11/numpy.h>
        template <int size, typename Basis>
        auto callGlobalIndexFromGlobalPosition(const Basis& basis, const Dune::FieldVector<double, size>& pos) {
            auto arrayOfIndices = Ikarus::utils::globalIndexFromGlobalPosition(basis,pos);
            auto InnerSize = arrayOfIndices[0].size(); // How many indices are there? 1 for flat basis
            auto OuterSize = arrayOfIndices.size(); // How many dofs are there at the poss

            auto x = pybind11::array_t<size_t>({OuterSize,InnerSize});
            auto r = x.template mutable_unchecked<2>();
            for (pybind11::ssize_t i = 0; i < r.shape(0); i++)
                for (pybind11::ssize_t j = 0; j < r.shape(1); j++)
                    r(i,j) = arrayOfIndices[i][j];
            return x;
        }
    """
    pos = FieldVector(pos)

    return run("callGlobalIndexFromGlobalPosition", StringIO(runCode), basis, pos)



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

# import debug_info
# debug_info.setDebugFlags()

import math

import ikarus as iks
from ikarus import finite_elements, assembler, io

import numpy as np
import scipy as sp

import dune.grid
import dune.functions

from dune.grid import gridFunction
from dune.vtk import FormatTypes, DataTypes

import os
import unittest
import pathlib

from dirichletvaluetest import makeGrid


class TestVtkWriter(unittest.TestCase):
    def setUp(self):
        reader = (
            dune.grid.reader.gmsh,
            os.path.join(os.path.dirname(__file__), "auxiliaryfiles/quad2d.msh"),
        )

        self.grid = dune.grid.ugGrid(reader, dimgrid=2)

        basis = iks.basis(
            self.grid, dune.functions.Power(dune.functions.Lagrange(order=1), 2)
        )
        self.flatBasis = basis.flat()
        d = np.zeros(len(self.flatBasis))

        lambdaLoad = iks.Scalar(3.0)

        def vL(x, lambdaVal):
            return np.array([lambdaVal * x[0] * 2, 2 * lambdaVal * x[1] * 0])

        vLoad = finite_elements.volumeLoad2D(vL)

        fes = []
        linElastic = finite_elements.linearElastic(youngs_modulus=1000, nu=0.2)
        for e in self.grid.elements:
            fes.append(finite_elements.makeFE(basis, linElastic, vLoad))
            fes[-1].bind(e)

        req = fes[0].createRequirement()
        req.insertParameter(lambdaLoad)
        req.insertGlobalSolution(d)

        self.dirichletValues = iks.dirichletValues(self.flatBasis)

        def fixLeftHandEdge(vec, localIndex, localView, intersection):
            if intersection.geometry.center[0] < 1e-8:
                vec[localView.index(localIndex)] = True

        self.dirichletValues.fixBoundaryDOFs(fixLeftHandEdge)

        self.sparseAssembler = assembler.sparseFlatAssembler(fes, self.dirichletValues)
        self.sparseAssembler.bind(
            req, iks.AffordanceCollection.elastoStatics, iks.DBCOption.Full
        )

        Msparse = self.sparseAssembler.matrix()
        forces = self.sparseAssembler.vector()

        self.x = sp.sparse.linalg.spsolve(Msparse, -forces)
        req.insertGlobalSolution(self.x)

    def assertIsFile(self, path):
        if not pathlib.Path(path).resolve().is_file():
            raise AssertionError("File does not exist: %s" % str(path))

    def test(self):
        _ = io.vtkWriter(
            self.sparseAssembler,
            datatype=DataTypes.Float64,
            headertype=DataTypes.UInt32,
        )

        writer = io.vtkWriter(self.sparseAssembler, dataFormat=FormatTypes.ascii)
        writer.addInterpolation(
            self.x, self.flatBasis, "displacements", io.DataTag.asPointData
        )
        writer.addInterpolation(
            self.x, self.flatBasis, "displacements", io.DataTag.asCellData
        )

        xDisplacementBasis = dune.functions.subspaceBasis(self.flatBasis, 0)
        writer.addInterpolation(self.x, xDisplacementBasis, "u", io.DataTag.asPointData)

        writer.addAllResults(io.DataTag.asCellData)
        writer.addAllResults(io.DataTag.asPointData)

        writer.write("vtk_test")
        self.assertIsFile("vtk_test.vtu")

    def test_lagrange(self):
        writer2 = io.vtkWriter(
            self.sparseAssembler, dataCollector=io.DataCollector.lagrange, order=2
        )

        writer2.addResult("linearStress", io.DataTag.asCellData)
        writer2.addResult("linearStress", io.DataTag.asPointData)

        writer2.setFormat(FormatTypes.ascii)
        writer2.setDatatype(DataTypes.Float64)

        @gridFunction(self.grid)
        def g(x):
            return [math.sin(2 * math.pi * x[0] * x[1]), x[0] * x[1]] * 5

        writer2.addPointData(g, name="g", components=(0, 1))
        writer2.addPointData(g, name="g2", components=[0, 1, 2])

        writer2.write("vtk_test_lagrange")

    def test_structuredWriter(self):
        gridUG = makeGrid()

        basis = iks.basis(
            gridUG, dune.functions.Power(dune.functions.Lagrange(order=1), 2)
        )
        flatBasisYASP = basis.flat()
        d = np.zeros(len(flatBasisYASP))

        lambdaLoad = iks.Scalar(3.0)

        fes = []
        linElastic = finite_elements.linearElastic(youngs_modulus=1000, nu=0.2)
        for e in gridUG.elements:
            fes.append(finite_elements.makeFE(basis, linElastic))
            fes[-1].bind(e)

        req = fes[0].createRequirement()
        lambdaLoad = iks.Scalar(3.0)
        req.insertParameter(lambdaLoad)
        req.insertGlobalSolution(d)

        dirichletValuesYASP = iks.dirichletValues(flatBasisYASP)
        sparseAssemblerYASP = assembler.sparseFlatAssembler(fes, dirichletValuesYASP)

        sparseAssemblerYASP.bind(
            req, iks.AffordanceCollection.elastoStatics, iks.DBCOption.Full
        )

        writer3 = io.vtkWriter(sparseAssemblerYASP, dataFormat=FormatTypes.ascii)
        writer3.addAllResults(io.DataTag.asCellData)

        writer3.write("vtk_test_structured")
        self.assertIsFile("vtk_test_structured.vtr")

        # self.assertEqual(fileName[-3:], "vtr")

    def test_discontinuousWriter(self):
        discontinuousVtkWriter = io.vtkWriter(
            self.sparseAssembler, dataCollector=io.DataCollector.discontinuous
        )
        discontinuousVtkWriter.addAllResults()  # defaults to pointData

        discontinuousVtkWriter.addInterpolation(
            self.x, self.flatBasis, "displacements", io.DataTag.asCellAndPointData
        )
        discontinuousVtkWriter.write("vtk_test_discontinous")


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import debug_info

debug_info.setDebugFlags()

import ikarus as iks
from ikarus import finite_elements, utils, assembler
import numpy as np
import scipy as sp
from scipy.optimize import minimize

import dune.grid
import dune.functions

if __name__ == "__main__":
    lowerLeft = []
    upperRight = []
    elements = []
    for i in range(2):
        lowerLeft.append(-1)
        upperRight.append(1)
        elements.append(3)

    grid = dune.grid.structuredGrid(lowerLeft, upperRight, elements)
    grid.hierarchicalGrid.globalRefine(0)
    basisLagrange1 = iks.basis(
        grid, dune.functions.Power(dune.functions.Lagrange(order=1), 2)
    )
    flatBasis = basisLagrange1.flat()

    forces = np.zeros(8)
    stiffness = np.zeros((8, 8))

    def vL(x, lambdaVal):
        return np.array([lambdaVal * x[0] * 2, 2 * lambdaVal * x[1] * 0])

    vLoad = iks.finite_elements.volumeLoad2D(vL)

    def neumannLoad(x, lambdaVal):
        return np.array([lambdaVal * 0, lambdaVal])

    neumannVertices = np.zeros(grid.size(2), dtype=bool)

    def loadTopEdgePredicate(x):
        return True if x[1] > 0.9 else False

    indexSet = grid.indexSet
    for v in grid.vertices:
        neumannVertices[indexSet.index(v)] = loadTopEdgePredicate(v.geometry.center)

    boundaryPatch = iks.utils.boundaryPatch(grid, neumannVertices)

    nBLoad = iks.finite_elements.neumannBoundaryLoad(boundaryPatch, neumannLoad)

    svk = iks.materials.StVenantKirchhoff(E=1000, nu=0.3)

    svkPS = svk.asPlaneStress()

    nonLinElastic = iks.finite_elements.nonLinearElastic(svkPS)

    fes = []
    for e in grid.elements:
        fes.append(
            iks.finite_elements.makeFE(basisLagrange1, nonLinElastic, vLoad, nBLoad)
        )
        fes[-1].bind(e)

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixFirstIndex(vec, globalIndex):
        vec[0] = True

    def fixAnotherVertex(vec, localIndex, localView):
        localView.index(localIndex)
        vec[1] = True

    def fixLeftHandEdge(vec, localIndex, localView, intersection):
        if intersection.geometry.center[1] < -0.9:
            vec[localView.index(localIndex)] = True

    dirichletValues.fixBoundaryDOFs(fixFirstIndex)
    dirichletValues.fixBoundaryDOFs(fixAnotherVertex)
    dirichletValues.fixBoundaryDOFs(fixLeftHandEdge)

    assembler = iks.assembler.sparseFlatAssembler(fes, dirichletValues)

    dRed = np.zeros(assembler.reducedSize())

    lambdaLoad = iks.Scalar(3.0)

    feReq = fes[0].createRequirement()

    def energy(dRedInput):
        feReq = fes[0].createRequirement()
        feReq.insertParameter(lambdaLoad)
        dBig = assembler.createFullVector(dRedInput)
        feReq.insertGlobalSolution(dBig)
        feReq.globalSolution()
        return assembler.scalar(feReq, iks.ScalarAffordance.mechanicalPotentialEnergy)

    def gradient(dRedInput):
        feReq = fes[0].createRequirement()
        feReq.insertParameter(lambdaLoad)
        dBig = assembler.createFullVector(dRedInput)
        feReq.insertGlobalSolution(dBig)
        return assembler.vector(
            feReq, iks.VectorAffordance.forces, iks.DBCOption.Reduced
        )

    def hess(dRedInput):
        feReq = fes[0].createRequirement()
        feReq.insertParameter(lambdaLoad)
        dBig = assembler.createFullVector(dRedInput)
        feReq.insertGlobalSolution(dBig)
        return assembler.matrix(
            feReq, iks.MatrixAffordance.stiffness, iks.DBCOption.Reduced
        ).todense()  # this is slow, but for this test we don't care

    resultd = minimize(energy, x0=dRed, options={"disp": True}, tol=1e-14)
    resultd2 = minimize(
        energy, x0=dRed, jac=gradient, options={"disp": True}, tol=1e-14
    )
    resultd3 = minimize(
        energy,
        method="trust-constr",
        x0=dRed,
        jac=gradient,
        hess=hess,
        options={"disp": True},
    )
    resultd4 = sp.optimize.root(gradient, jac=hess, x0=dRed, tol=1e-10)

    assert np.allclose(resultd.x, resultd2.x, atol=1e-6)
    assert np.allclose(resultd3.x, resultd4.x)
    assert np.all(abs(resultd3.grad) < 1e-8)
    assert np.all(abs(resultd4.fun) < 1e-8)

    feReq = fes[0].createRequirement()
    fullD = assembler.createFullVector(resultd2.x)
    feReq.insertGlobalSolution(fullD)

    res1 = fes[0].calculateAt(feReq, np.array([0.5, 0.5]), "PK2Stress")



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt

from ikarus.generator import MySimpleGenerator
from dune.vtk import FormatTypes, DataTypes
import dune.vtk

import types
from enum import Enum

from ._io import DataTag

# The list of supported dataCollectors
DataCollector = Enum("DataCollector", ["default", "lagrange", "discontinuous", "iga"])


def __addInterpolation(writer):
    def __addInterpolationFunc(
        writer, vals_, basis, name: str, dataTag=DataTag.asPointData
    ):
        gf = basis.asFunction(vals_)
        if dataTag == DataTag.asPointData or DataTag.asCellAndPointData:
            writer.addPointData(gf, name)
        if dataTag == DataTag.asCellData or DataTag.asCellAndPointData:
            writer.addCellData(gf, name)

    return __addInterpolationFunc


def vtkWriter(
    assembler,
    dataCollector: DataCollector = DataCollector.default,
    order: int = 1,
    dataFormat=FormatTypes.binary,
    datatype=DataTypes.Float32,
    headertype=DataTypes.UInt32,
):
    includes = []
    includes += ["dune/python/vtk/writer.hh"]
    includes += ["dune/vtk/writers/unstructuredgridwriter.hh"]
    includes += ["dune/vtk/writers/unstructuredgridwriter.hh"]
    includes += ["dune/vtk/datacollectors/lagrangedatacollector.hh"]
    includes += ["ikarus/io/vtkwriter.hh"]
    includes += ["ikarus/assembler/simpleassemblers.hh"]
    includes += assembler._includes

    gridViewName = assembler.gridView.cppTypeName
    dataCollectorName: str = ""

    if dataCollector is not DataCollector.default:
        if dataCollector == DataCollector.lagrange:
            dataCollectorName = (
                f"Dune::Vtk::LagrangeDataCollector<{gridViewName}, {order}>"
            )
        elif dataCollector == DataCollector.discontinuous:
            dataCollectorName = f"Dune::Vtk::DiscontinuousDataCollector<{gridViewName}>"
            includes += ["dune/vtk/datacollectors/discontinuousdatacollector.hh"]
        elif dataCollector == DataCollector.iga:
            dataCollectorName = (
                f"Dune::Vtk::DiscontinuousIgaDataCollector<{gridViewName}>"
            )
            includes += ["dune/iga/io/igadatacollector.hh"]

    generator = MySimpleGenerator("VtkWriter", "Ikarus::Python")

    defaultManager = f"Ikarus::Vtk::DefaultVTKWriterManager<{gridViewName}>"
    vtkWriterName = ""
    element_type = ""

    if dataCollector is DataCollector.default:
        vtkWriterName = f"typename {defaultManager}::template DefaultVTKWriter<>"
        dataCollectorName = f"typename {defaultManager}::DefaultDataCollector"
        element_type = f"Ikarus::Vtk::Writer<{assembler.cppTypeName}, {dataCollectorName}, {vtkWriterName}>"

    else:
        vtkWriterName = (
            f"typename {defaultManager}::template DefaultVTKWriter<{dataCollectorName}>"
        )
        element_type = f"Ikarus::Vtk::Writer<{assembler.cppTypeName}, {dataCollectorName}, {vtkWriterName}>"

    # Register VTKWriter
    dune.vtk.load(includes, vtkWriterName)

    includes += ["ikarus/python/io/vtkwriter.hh"]
    moduleName = "vtkWriter_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName,
        baseClasses=[vtkWriterName],
        dynamicAttr=True,
    )
    writerModule = module.VtkWriter(assembler, dataFormat, datatype, headertype)
    writerModule.addInterpolation = types.MethodType(
        __addInterpolation(writerModule), writerModule
    )

    return writerModule
